#include "arguments.h"
#include "csv.h"
#include "utility/utility.h"

using namespace kaminpar;
using namespace kaminpar::tool;

std::vector<std::pair<std::string, std::string>> parse_aggregates(const std::string &aggregates) {
  std::vector<std::pair<std::string, std::string>> result;
  for (const auto &entry : utility::str::explode(aggregates, ',')) {
    const auto entries = utility::str::explode(entry, ':');
    if (entries.size() != 2) { FATAL_ERROR << "invalid aggregate format: " << entry; }
    result.push_back({entries[0], entries[1]});
  }
  return result;
}

int main(int argc, char *argv[]) {
  std::vector<std::string> filenames{};
  std::string group_by{};
  std::string aggregates{};
  bool colorize{true};
  std::size_t expected_rows_per_aggregate{0};
  bool filter_failed{false};
  bool quiet{false};

  // clang-format off
  Arguments args;
  args.positional().argument("files", "One or multiple *.csv files", &filenames);
  args.group("Options")
      .argument("group-by", "Group by columns", &group_by, 'G')
      .argument("aggregate", "Aggregate function for each column, e.g., ColumnA:avg,ColumnB:sum,ColumnC:count", &aggregates)
      .argument("colorize", "Colorize output", &colorize, 'C')
      .argument("filter-failed", "Filter failed runs", &filter_failed, 'f')
      .argument("quiet", "Do not produce additional output", &quiet, 'q')
      .argument("expected-rows-per-aggregate", "Warn if the number of rows per aggregate differs for some entries (0: don't warn)", &expected_rows_per_aggregate);
  args.parse(argc, argv);
  // clang-format on

  if (filenames.empty()) { FATAL_ERROR << "Need at least one *.csv file to work with"; }
  if (group_by.empty()) { FATAL_ERROR << "Need at least one column to group"; }

  std::vector<Csv> csvs;
  for (const std::string &filename : filenames) { csvs.push_back(load_csv(filename, filter_failed, ',', quiet)); }

  const auto group_by_columns = utility::str::explode(group_by, ',');
  const auto aggregate_pairs = parse_aggregates(aggregates);

  // find all columns that we want to keep
  std::vector<std::string> common_columns{group_by_columns};
  for (const auto &[column, function] : aggregate_pairs) {
    if (std::find(common_columns.begin(), common_columns.end(), column) != common_columns.end()) {
      FATAL_ERROR << column << " appears in both the aggregates and group by list";
    }
    common_columns.push_back(column);
  }

  // collapse csvs into a big one
  Csv unaggregated_csv = merge_csvs(common_columns, csvs);
  Csv aggregated_csv = aggregate(unaggregated_csv, group_by_columns, aggregate_pairs, expected_rows_per_aggregate);

  print_csv(aggregated_csv, colorize);
  return 0;
}